/* Terraform - (C) 1997-1999 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite s_HBOffs_HBOff0,
 * Boston, MA 02111-1s_HBOff07, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TFOptions.h"
#include "TFFileRC.h"
#include "HeightFieldGenRandom.h"
#include "GlobalTrace.h"
#include "terraform.h"				// defines version number

extern int readhf (HeightField **HF, char *fname, bool haveGUI);

// defined in TFWindow.cc
extern guint TFWhash (gconstpointer key);
extern gint TFWhashcmp (gconstpointer a, gconstpointer b);

// defined in TFFileRC.cc
void hash_key_fetcher (gpointer key, gpointer data, gpointer user_data);
void hash_printer (gpointer key, gpointer data, gpointer user_data);
void hash_key_printer (gpointer key, gpointer data, gpointer user_data);



/*
 *  processArgs: process all args 
 */
void TFOptions::processArgs (HeightField **HF,int argc, char **argv)
{
	processGlobalArgs (argc, argv);
	processAppArgs (HF, argc, argv);
	processRCFile ();
}


/*
 *  processAppArgs: process the application-related command line arguments 
 */
void TFOptions::processAppArgs (HeightField **HF,int argc, char **argv)
{
	int			rc;

	for (int i=1; i<argc; i++)
		{
		if (argv[i] &&
	            (!strcmp(argv[i],"-f")||!strcmp(argv[i],"--file")) )
			{
			if (argc >= i+1 && argv[i+1])
				{
				if ((rc=readhf (HF, argv[++i], FALSE)) < 0)
					{
					printf ("File read failed (rc=%d) for file [%s] ... Exiting\n", rc, argv[i]);
					exit (-1);
					}
				}
			else
				{
				printf ("Error: file argument for %s <file> missing ... Exiting\n", argv[i]);
				exit (-1);
				}
			}
		else
		if (!strcmp (argv[i], "-g") || !strcmp (argv[i], "--generate"))
			{
			*HF = new HeightField ();
			HeightFieldGenRandom::generate (*HF, 400);
			}
#ifdef NEVER
		else
		if (!strcmp (argv[i], "-cli"))
			do_gui=FALSE;
#endif
		else 

		// dummy checks we do in process_global_args
		if (!strcmp (argv[i], "-dbg") || !strcmp (argv[i], "--debug"))
			i++;
		else
		if (!strcmp (argv[i], "-v") || !strcmp (argv[i], "--verbose"))
			;
		else
			{
			printf ("Error: argument [%s] not recognized ... \n", 
				argv[i]);
			help ();
			exit (0);
			}
		}
} 


/*
 *  processGlobalArgs: process the global command line arguments 
 */
void TFOptions::processGlobalArgs (int argc, char **argv)
{
	// see if we should set a certain trace level 
	for (int i=1; i<argc; i++)
		{
		if (!strcmp (argv[i], "-dbg") || !strcmp (argv[i], "--debug"))
			{
			if (argc >= i+2)	// 1 for count-to-index, 1 for next arg
				GlobalTrace::setTraceLevel (atoi (argv[++i]));
			else
				{
				printf ("Error: traceLevel argument missing for %s ... Exiting\n", argv[i]);
				exit (1);
				}
			}
		else
		if (!strcmp (argv[i], "-v") || !strcmp (argv[i], "--verbose"))
			GlobalTrace::setTraceLevel (GlobalTrace::TRACE_VERBOSE);
		else
		if (!strcmp (argv[i], "-version") || !strcmp(argv[i], "--version"))
			{
			printf ("terraform %d.%d.%d\n", TF_MAJOR_VERSION, 
				TF_MINOR_VERSION, TF_MICRO_VERSION);
			exit (0);
			}
		else
		if (!strcmp (argv[i], "--help") || !strcmp (argv[i], "-?"))
			{
			help ();
			exit (0);
			}
		else
		if (!strcmp (argv[i], "--helprc") || !strcmp (argv[i], "-hrc"))
			{
			helpRCFile ();
			exit (0);
			}
		else
		if (!strcmp (argv[i], "--installrc") || !strcmp (argv[i], "-irc"))
			{
			install ();
			}

		}
}

void TFOptions::install ()
{
	TFFileRC	*rcFile = NULL;
	char		*home, 
			buf[80],
			buf2[80];

	home = getenv ("HOME");
	SanityCheck::warning((!home), "can't determine $HOME, cant create RC file", "TFOptions::processRCFile");
	if (home)
		{
		bool	overwrite=TRUE;

		sprintf (buf, "%s/%s", home, TF_LOCAL_RC_FILE);
		rcFile = new TFFileRC (buf, NULL);
		if (rcFile->exists())
			overwrite = rcFile->promptOverwrite ();
		if (overwrite)
			{
			sprintf (buf2, "Creating %s ... ", buf);
			GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf2);
			rcFile->open ("w");
			if (rcFile->getFilePtr())
				{
				updateOptionTable ();
				rcFile->writeRCFile (this);
				}
			rcFile->close ();
			GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "done\n");
			}
		delete rcFile;
		}

	exit (0);
}


/*
 *  processRCFile: process the resource file
 */
void TFOptions::processRCFile ()
{
	TFFileRC	*rcFile;
	TFOptionsTable	*optTbl=NULL;
	char		*buf[4], 
			*home, 
			t[128];
	int		lim = 0,
			c = 0;

	sprintf (t, "%s/%s", TF_DATADIR, TF_SYSTEM_RC_FILE);
	buf[lim++] = strdup (t);

	home = getenv ("HOME");
	SanityCheck::warning((!home), "can't determine $HOME", "TFOptions::processRCFile");
	if (home)
		{
		sprintf (t, "%s/%s", home, TF_LOCAL_RC_FILE);
		buf[lim++] = strdup (t);
		}

	while (c < lim)
		{
		rcFile = new TFFileRC (buf[c++], optTbl);
		if (rcFile->exists())
			optTbl = rcFile->parseRCFile ();
		delete rcFile;
		}

	if (optTbl)
		updateTFOptions (optTbl);
	else
		GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "Couldn't find any resource files\n");
		

}


/*
 * updateTFOptions: write the options in optTbl to the actual TFOptions variables
 */
void TFOptions::updateTFOptions (TFOptionsTable *optTbl)
{
	gpointer	t;
	char 		buf[80];
	int		changed=0;

	t = optTbl->lookup (TF_OPT_FILL_SEA);
	if (t)
		{
		bool	b = (bool) atoi ((const char*) t);
		TFOptions::s_fillSea = b;
		changed++;
		}

	t = optTbl->lookup (TF_OPT_DRAW_DERIVED_DATA);
	if (t)
		{
		bool	b = (bool) atoi ((const char*) t);
		TFOptions::s_drawDerivedData = b;
		changed++;
		}

	t = optTbl->lookup (TF_OPT_FAST_HALF_RES);
	if (t)
		{
		bool	b = (bool) atoi ((const char*) t);
		TFOptions::s_fastHalfRes = b;
		changed++;
		}

	t = optTbl->lookup (TF_OPT_FAST_HALF_Y_SCALE);
	if (t)
		{
		bool	b = (bool) atoi ((const char*) t);
		TFOptions::s_fastHalfYscale = b;
		changed++;
		}

	t = optTbl->lookup (TF_OPT_FAST_ZROT);
	if (t)
		{
		bool	b = (bool) atoi ((const char*) t);
		TFOptions::s_fastZrot = b;
		changed++;
		}

	t = optTbl->lookup (TF_OPT_SCALE);
	if (t)
		{
		int	i = atoi ((const char*) t);
		TFOptions::s_scale = i;
		changed++;
		}

	t = optTbl->lookup (TF_OPT_SYNC_RATE);
	if (t)
		{
		int	i = atoi ((const char*) t);
		TFOptions::s_syncRate = i;
		changed++;
		}

	t = optTbl->lookup (TF_OPT_RESOLUTION);
	if (t)
		{
		int	i = atoi ((const char*) t);
		TFOptions::s_resolution = i;
		changed++;
		}

	t = optTbl->lookup (TF_OPT_YSCALE);
	if (t)
		{
		float	f = (float) atof ((const char*) t);
		TFOptions::s_yscale = f;
		changed++;
		}

	t = optTbl->lookup (TF_OPT_SEALEVEL);
	if (t)
		{
		float	f = (float) atof ((const char*) t);
		TFOptions::s_sealevel = f;
		changed++;
		}
	t = optTbl->lookup (TF_OPT_POV_FILE);
	if (t)
		{
		char *s = (char *) t;
		strcpy (TFOptions::s_POVfile, s);
		changed++;
		}
	t = optTbl->lookup (TF_OPT_POV_WIDTH);
	if (t)
		{
		int	i = (int) atoi ((const char*) t);
		TFOptions::s_POVwidth = i;
		changed++;
		}
	t = optTbl->lookup (TF_OPT_POV_YSCALE);
	if (t)
		{
		bool 	b = (bool) atoi ((const char*) t);
		TFOptions::s_POVhalfYscale = b;
		changed++;
		}
	t = optTbl->lookup (TF_OPT_POV_KEEP_FILES);
	if (t)
		{
		bool 	b = (bool) atoi ((const char*) t);
		TFOptions::s_POVkeepFiles = b;
		changed++;
		}

	sprintf (buf, "Updated %d params from OptionsTable to TFOptions\n", changed);
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, buf);
}


/*
 * updateOptionTable: update the optionTable from TFOptions
 */
void TFOptions::updateOptionTable ()
{
	char 	buf[80];

	this->insert (TF_OPT_FILL_SEA, (s_fillSea ? "1" : "0"));
	this->insert (TF_OPT_DRAW_DERIVED_DATA, (s_drawDerivedData ? "1" : "0"));
	this->insert (TF_OPT_FAST_HALF_RES, (s_fastHalfRes ? "1" : "0"));
	this->insert (TF_OPT_FAST_HALF_Y_SCALE, (s_fastHalfYscale ? "1" : "0"));
	this->insert (TF_OPT_FAST_ZROT, (s_fastZrot ? "1" : "0"));
	this->insert (TF_OPT_POV_YSCALE, (s_POVhalfYscale ? "1" : "0"));
	this->insert (TF_OPT_POV_KEEP_FILES, (s_POVkeepFiles ? "1" : "0"));

	sprintf (buf, "%d", s_scale);
	this->insert (TF_OPT_SCALE, strdup(buf));
	sprintf (buf, "%d", s_syncRate);
	this->insert (TF_OPT_SYNC_RATE, strdup(buf));
	sprintf (buf, "%d", s_resolution);
	this->insert (TF_OPT_RESOLUTION, strdup(buf));
	sprintf (buf, "%d", s_POVwidth);
	this->insert (TF_OPT_POV_WIDTH, strdup(buf));

	sprintf (buf, "%.2f", s_sealevel);
	this->insert (TF_OPT_SEALEVEL, strdup(buf));
	sprintf (buf, "%.2f", s_yscale);
	this->insert (TF_OPT_YSCALE, strdup(buf));

	sprintf (buf, "%s/%s", TF_DATADIR, TF_DEFAULT_POV_FILE);
	this->insert (TF_OPT_POV_FILE, strdup(buf));
}


/*
 * helpRCFile: display the recognized flags for the RC file
 */
void TFOptions::helpRCFile ()
{
	printf ("terraform %d.%d.%d Resource File Info:\n", TF_MAJOR_VERSION, 
				TF_MINOR_VERSION, TF_MICRO_VERSION);
	printf ("System data dir: %s, datafile: %s\n", TF_DATADIR, TF_SYSTEM_RC_FILE);
	printf ("User resorce file: $HOME/%s\n", TF_LOCAL_RC_FILE);
	printf ("\nRecognized terraform resource file tokens are:\n");

	FlexArray	*flArr = this->collectOptions ();

	for (int i=0; i<flArr->getSize(); i++)
		{
		char *k = (char *)(flArr->El(i));
		hash_key_printer (k, NULL, stdout);
		}
}


/*
 * help: display recognized command line options 
 */
void TFOptions::help ()
{
	printf ("terraform %d.%d.%d\n", TF_MAJOR_VERSION, TF_MINOR_VERSION, 
		TF_MICRO_VERSION);
	printf ("  -dbg, --debug <level>		set debug to the specified level\n");
	printf ("  -f, --file <file>		read Height Field <file> upon startup\n");
	printf ("  -hrc, --helprc 		print resource file help\n");
	printf ("  -g, --generate 		generate a Height Field upon startup\n");
	printf ("  -irc, --installrc		install/create user resource file and exit\n");
	printf ("  -version, --version		print version\n");
	printf ("  -v, --verbose			verbose mode\n");
	printf ("  -?, --help			print help and exit\n");
	printf ("\n");
}



void hash_key_fetcher (gpointer key, gpointer data, gpointer user_data)
{
	char		*k = (char*)key;
	FlexArray	*f = (FlexArray*)user_data;

	f->append (k);
}


void hash_printer (gpointer key, gpointer data, gpointer user_data)
{
	char	*k = (char*)key,
		*d = (char*)data;
	FILE	*f = (FILE*)user_data;

	fprintf (f, "%s=%s\n", k, d);
}


void hash_key_printer (gpointer key, gpointer data, gpointer user_data)
{
	char	*k = (char*)key;
	FILE	*f = (FILE*)user_data;

	fprintf (f, "\t%s\n", k);
}



